/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2010 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>

#include <tracefct/tracefct.h>
#include <linklist/linklist.h>

#include "luav.h"

#include "aperture.h"
#include "lua_register.h"

#include "luasup.h"

const char*
get_opt_label(
	      lua_State *L,
	      int *narg
	       )
{
    const char* label = NULL;

    if ( lua_gettop( L ) > *narg )
    {
	(*narg)++;
	luaL_checktype( L, *narg, LUA_TTABLE );
	lua_getfield( L, *narg, APERTURELABEL );

	if ( lua_isstring( L, -1 ) )
	{
	    label = lua_tostring( L, -1 );
	}
	else if ( ! lua_isnil( L, -1 ) )
	{
	    luaL_error( L, "bad argument '%s' (%s expected, got %s)",
			APERTURELABEL, lua_typename( L, LUA_TSTRING), luaL_typename( L, -1 ) );
	}

	lua_pop( L, 1 );
    }

    return label;
}


void stackDump( lua_State *L )
{
    int i;
    int top = lua_gettop(L);

    for ( i = 1; i <= top ; i++ )
    {
	int t = lua_type(L, i );
	switch( t )
	{
	case LUA_TSTRING:
	    fprintf( stderr, "%d: string: '%s'\n", i, lua_tostring(L, i ));
	    break;

	case LUA_TBOOLEAN:
	    fprintf( stderr, "%d: boolean: %s\n", i, lua_toboolean(L,i) ? "true" : "false");
	    break;

	case LUA_TNUMBER:
	    fprintf( stderr, "%d: number: %g\n", i, lua_tonumber(L, i));
	    break;

	default:
	    fprintf( stderr, "%d: type: %s\n", i, lua_typename(L,t));
	    break;
	}
    }
    fprintf( stderr, "\n");
}

int
checkboolean( lua_State *L, int narg ) {

    if ( !lua_isboolean( L, narg ) )
	return luaL_typerror( L, narg, lua_typename(L, LUA_TBOOLEAN) );

    return lua_toboolean( L, narg );
}


static int
lualib_cp_node_to_array( void *node, void *udata )
{
    luaL_Reg *from = node;
    luaL_Reg **top = udata;
    luaL_Reg *to   = *top;

    *to = *from;
    (*top)++;

    return 0;
}

static
void ufree( void * p)
{
    free( p );
}

int
luaopen_aperture( lua_State *L )
{
    luaL_Reg *mylib;
    luaL_Reg *to_p;

    SLinkList liblist;

    tf_enter( "luaopen_aperture" );

    liblist = sll_new( SLL_NULL_CMP );
    if ( NULL == liblist )
	tf_die( "error allocating luaL_Reg linked list" );

    /* dynamically created list of functions */
    {
	ApertureModule **mod;
	for (mod = modules ; *mod ; mod++ )
	    register_luacfunc( liblist, (*mod)->name, (*mod)->function );
    }

    /* register other systems */
    assembly_queue_register_lua( liblist );
    transform_register_lua( liblist );
    xstack_register_lua( liblist );

    /* now create a real array */

    to_p = mylib = malloc( sizeof( luaL_Reg ) * (sll_count( liblist ) + 1 ) );
    if ( NULL == mylib )
	tf_die( "error allocating luaL_Reg list" );

    sll_utraverse( liblist, lualib_cp_node_to_array, &to_p );

    to_p->name = NULL;
    to_p->func = NULL;

    lua_getglobal( L, "require" );
    lua_pushstring( L, "saotrace.aperture" );
    if ( lua_pcall( L, 1, 0, 0) )
        tf_die( lua_tostring(L, -1 ) );

    luaL_register( L, "saotrace.aperture", mylib );

    sll_delete( liblist, ufree );

    free( mylib );

    tf_leave();

    return 1;
}


void
register_luacfunc( SLinkList liblist, const char *name, lua_CFunction func )
{
    luaL_Reg *l = malloc( sizeof(luaL_Reg) );

    if ( NULL ==  l )
	tf_die( "unable to allocate luaL_Reg struct" );

    l->name = name;
    l->func = func;

    sll_insert_tail( liblist, l );
}
