// --8<--8<--8<--8<--
//
// Copyright (C) 2011 Smithsonian Astrophysical Observatory
//
// This file is part of bpmanip
//
// bpmanip is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

#include <iostream>

#include <tracefctxx/TraceFct.h>

#include "BinaryPipe.h"

using std::string;

BinaryPipe::BinaryPipe ( string& input, string& output ) : m_packet_data_ptr ( NULL )
{
    TraceFct tf( "BinaryPipe::BinaryPipe" );

    // bpipe_errno is a global in the library that is set
    //   when an error occurs.
    //
    m_bpipe = bpipe_new ();

    if ( 0 != bpipe_errno )
	tf.die( "binary pipe init failure: %s", bpipe_strerror( bpipe_errno ) );

    int retval = bpipe_input  ( m_bpipe, input.c_str() );

    if ( 0 != bpipe_errno || 0 != retval )
	tf.die( "binary pipe input init failure: %s", bpipe_strerror( bpipe_errno ) );

    m_bpipe_out       = bpipe_output ( m_bpipe, output.c_str()  );

    if ( 0 != bpipe_errno )
	tf.die( "binary pipe output init failure: %s", bpipe_strerror( bpipe_errno ) );
}

BinaryPipe::~BinaryPipe () {

    // Be careful the m_bpipe hasn't already been deleted.
    //
    if ( NULL != m_bpipe ) {
        bpipe_delete ( m_bpipe ); // library function to delete BPipe
        m_bpipe           = NULL;
        m_bpipe_out       = NULL;
        m_packet_data_ptr = NULL;
    }
}

void BinaryPipe::WriteHeader () {

    TraceFct tf( "BinaryPipe::WriteHeader" );

    // bpipe_write_hdr is a library function that output
    //   the bp header to the set output.
    //
    bool retval = ( 0 == bpipe_write_hdr ( m_bpipe ) );

    if ( !retval && 0 != bpipe_errno )
	tf.die( "error writing binary pipe header: %s", bpipe_strerror( bpipe_errno ) );
}

void BinaryPipe::MapAlloc ( size_t n ) {

    TraceFct tf( "BinaryPipe::MapAlloc" );

    // bpipe_map_alloc (library function) freezes the output format
    //   after any changes have been made to the format.  It must
    //   be called before anything is written, either header or data packet.
    // It returns NULL on error, and sets bpipe_errno.
    //
    m_packet_data_ptr = bpipe_map_alloc ( m_bpipe, n, NULL );

    if ( 0 != bpipe_errno || NULL == m_packet_data_ptr )
	tf.die( "map allocation error: %s", bpipe_strerror( bpipe_errno ) );
}

void BinaryPipe::DeleteDataField ( const string& filename ) {

    // bpipe_dpktf_delete (library function) deletes data fields from
    //   the core image, which in turn deletes them from the output
    //   format.  This must be done BEFORE anything has been written
    //   to output, or Bad Things will happen.  A good piece of added
    //   functionality would be to check against that.
    //
    bpipe_dpktf_delete ( m_bpipe, filename.c_str(), BPDSite_CORE, m_bpipe_out );
}

void BinaryPipe::GetDataFieldNames ( Container& fieldnames ) {

    // Get each data field, one by one.  The library routine bpipe_dpktf_next
    //   gets the next data packet field and returns NULL when there are no
    //   more fields.  bpipe_dpktf_name (library function) gets the name of
    //   the field pointed to by data_pkt.
    //
    void*      m_last = NULL;
    DpktField* data_pkt = NULL; // library type

    while ( NULL != ( data_pkt = bpipe_dpktf_next ( m_bpipe, &m_last ) ) )
        fieldnames.push_back ( bpipe_dpktf_name ( data_pkt ) );
}

void BinaryPipe::AlterCore  ( Fields& allowed_fields ) {

    Container input_fields;
    GetDataFieldNames( input_fields );

    Container deleted_fields = allowed_fields.rejected( input_fields );

    for ( ConIter field = deleted_fields.begin() ;
	  field < deleted_fields.end() ;
	  field++ )
	DeleteDataField ( *field );

    MapAlloc( buffer_size );
}

void BinaryPipe::CopyPackets( ) {

    TraceFct tf( "BinaryPipe::CopyPackets" );

    // Read each packet in the input file in, one by one, and print it
    //   out, one by one.  Since the core image has been changed by
    //   data_pkt.AlterCore, the output will be what the user requested.
    //
    int packet_count = 0;
    int nread;
    while ( nread = bpipe_read_dpkts ( m_bpipe, m_packet_data_ptr, buffer_size) ) {

	if ( 0 != bpipe_write_dpkts (m_bpipe, m_packet_data_ptr, nread, m_bpipe_out) )
	    tf.die( "error writing packet %d\n", packet_count );

	++packet_count;

    }

    if ( 0 != bpipe_errno )
	tf.die( "error reading input stream: %s", bpipe_strerror( bpipe_errno ) );
}
