#include <string>
#include <sstream>

#include <tracefctxx/TraceFct.h>
#include <Exception/Exception.h>

#include "units.h"
#include "utils.h"

#include <Eigen/SVD>
using namespace Eigen;

using namespace raygen;

#include "Jitter/Table.h"

namespace Jitter {

    //---------------------------------------------------------------------

    void
    Table::read( TableFile::Base* file ) {

	_size = file->read( points, tvalid );

	// if input quaternions are additive, make them absolute
	// relative to position of optical axis
	if ( rotation_is_additive ) {

	    Quaterniond initial = optical_axis.inverse() * input_origin ;

	    for_each( points, points + _size, AdditiveToAbsolute(initial) );
	}

	// if input quaternions are non-additive, make them relative
	// to position of optical axis
	else {

	    for_each( points, points + _size,
		      ApplyPostQuat(optical_axis.inverse()) );
	}

    }

    //---------------------------------------------------------------------

    void
    Table::jitter( double time, Quaterniond& q ) {

	if ( time < tvalid.min || time >= tvalid.max ) {

	    switch( _oob ) {

	    case OOB_Zero:
		q = Eigen::Quaterniond::Identity();
		return;
		break;

	    case OOB_Throw:
		{
		    ostringstream str;
		    str << "requested time ("
			<< time
			<< ") is out of range of jitter table ("
			<< points[0].time << ", "
			<< points[size()-1].time
			<< ")\n";
		    throw( Exception( str.str() ) );
		    break;
		}

	    default:
		TraceFct::die( "internal error: unknown Time OutOfBounds method: %d", _oob );
	    }

	}

	// find bounding jitter points in time.  "typically" the
	// interval in time between jitter points is on the order of,
	// or larger than, the interval in time between rays, so we
	// can do a linear search from the last jitter point and not
	// be too inefficient.

	// determine the index of the jitter entry which is minimally
	// greater than the requested time
	size_t pidx;
	for (  pidx = _last.idx; pidx < size() && time >= points[pidx].time ; pidx++ )
	    ;

	tf_assert( pidx > 0 && pidx <= size() );

	// convert to start of interval. if pidx == size the ray time is outside of the
	// jitter table, so extrapolate
	if ( pidx == size() ) pidx--;
	pidx--;

	// calculate current rotation by interpolation between
	// points[pidx] and points[pidx+1],  assuming that the change in
	// rotation is constant over the jitter period.
	double dt = (time - points[pidx].time) / ( points[pidx+1].time - points[pidx].time );


	Quaterniond now = points[pidx].q.slerp( dt, points[pidx+1].q );

#if 0
	std::cerr << "pidx = " << pidx
		  << "\n"
		  << "dt = " << dt
		  << "\n"
		  << "q0 = " << points[pidx].q
		  << "\n"
		  << "qn = " << now
		  << "\n"
		  << "q1 = " << points[pidx+1].q
		  << "\n"
	    ;

	std::cerr << "ql = " << _last.q << "\n";

#endif
	// now get delta from last rotation to current rotation
	q = now * _last.q.inverse();

	_last.idx = pidx;
	_last.q = now;

    }


}
