module( ..., package.seeall )

local table = require( 'table' )
local os = require( 'os' )

local deepcompare = require( 'deepcompare' )

local srcdir = os.getenv('srcdir') or '.'

local ea_db = table.concat( { srcdir, 'data', 'annulus.rdb'}, '/' )

local utils = require( 'saotrace.raygen.ea.utils.ring_n_spoke' )
local va = require('validate.args')
local validate = va.validate

function test_spokes_wrap_single( )

   local pars = { rmin = 1,
		  rmax = 1,
		  nspokes = 1,
		  nrings = 1,
		  tmin = 0,
		  tmax = 360,
		  dist = 'uniform',
	       }

   local spoke = 1
   for spot in utils.ringNspoke( pars ) do

      local exp = pars.tmin;
      assert_equal( exp, spot.theta, "spoke " .. spoke )

      spoke = spoke + 1
   end

   assert_equal( 2, spoke )

end

function test_spokes_wrap_odd( )

   local pars = { rmin = 1,
		  rmax = 1,
		  nspokes = 5,
		  nrings = 1,
		  tmin = 0,
		  tmax = 360
	       }

   local spoke = 1
   for spot in utils.ringNspoke( pars ) do

      local exp = pars.tmin + ( spoke - 1 ) * ( pars.tmax - pars.tmin ) / pars.nspokes
      assert_equal( exp, spot.theta, "spoke " .. spoke )

      spoke = spoke + 1
   end

end

function test_spokes_wrap_even( )

   local pars = { rmin = 1,
		  rmax = 1,
		  nspokes = 4,
		  nrings = 1,
		  tmin = 0,
		  tmax = 360
	       }

   local spoke = 1
   for spot in utils.ringNspoke( pars ) do

      local exp = pars.tmin + ( spoke - 1 ) * ( pars.tmax - pars.tmin ) / pars.nspokes
      assert_equal( exp, spot.theta, "spoke " .. spoke )

      spoke = spoke + 1
   end

end

function test_spokes_bounded_single( )

   local pars = { rmin = 1,
		  rmax = 1,
		  nspokes = 1,
		  nrings = 1,
		  tmin = 0,
		  tmax = 75
	       }

   local spoke = 1
   for spot in utils.ringNspoke( pars ) do

      local exp = (pars.tmin + pars.tmax ) / 2
      assert_equal( exp, spot.theta, "spoke " .. spoke )

      spoke = spoke + 1
   end

   assert_equal( 2, spoke )

end

function test_spokes_bounded_odd( )

   local pars = { rmin = 1,
		  rmax = 1,
		  nspokes = 5,
		  nrings = 1,
		  tmin = 0,
		  tmax = 75
	       }

   local spoke = 1
   local last_spot
   for spot in utils.ringNspoke( pars ) do

      local exp = pars.tmin + ( spoke - 1 ) * ( pars.tmax - pars.tmin ) / ( pars.nspokes - 1 )
      assert_equal( exp, spot.theta, "spoke " .. spoke )

      spoke = spoke + 1
      last_spot = spot
   end

   assert_equal( pars.tmax, last_spot.theta )

end


function test_spokes_bounded_even( )

   local pars = { rmin = 1,
		  rmax = 1,
		  nspokes = 4,
		  nrings = 1,
		  tmin = 0,
		  tmax = 75
	       }

   local last_spot
   local spoke = 1
   for spot in utils.ringNspoke( pars ) do

      local exp = pars.tmin + ( spoke - 1 ) * ( pars.tmax - pars.tmin ) / ( pars.nspokes - 1 )
      assert_equal( exp, spot.theta, "spoke " .. spoke )

      spoke = spoke + 1
      last_spot = spot
   end

   assert_equal( pars.tmax, last_spot.theta )

end


function test_rings_single( )

   local pars = { rmin = 1,
		  rmax = 10,
		  nspokes = 1,
		  nrings = 1,
		  tmin = 0,
		  tmax = 75,
		  dist = 'uniform',
	       }

   local ring = 1
   for spot in utils.ringNspoke( pars ) do

      local exp = (pars.rmin + pars.rmax ) / 2
      assert_equal( exp, spot.r, "ring " .. ring )

      ring = ring + 1
   end

   assert_equal( 2, ring )

end

function test_rings_double( )

   local pars = { rmin = 1,
		  rmax = 10,
		  nspokes = 1,
		  nrings = 2,
		  dist = 'uniform'
	       }

   local exp = { pars.rmin, pars.rmax }

   local ring = 1
   for spot in utils.ringNspoke( pars ) do

      assert_equal( exp[ring], spot.r )

      ring = ring + 1
   end

   assert_equal( 3, ring )

end


function test_rings_odd( )

   local pars = { rmin = 1,
		  rmax = 10,
		  nspokes = 1,
		  nrings = 5,
		  tmin = 0,
		  tmax = 75,
		  dist = 'uniform'
	       }

   local ring = 1
   local last_spot
   for spot in utils.ringNspoke( pars ) do

      local exp = pars.rmin + ( ring - 1 ) * ( pars.rmax - pars.rmin ) / ( pars.nrings - 1 )
      assert_equal( exp, spot.r, "ring " .. ring )

      ring = ring + 1
      last_spot = spot
   end

   assert_equal( pars.rmax, last_spot.r )

end


function test_rings_even( )

   local pars = { rmin = 1,
		  rmax = 10,
		  nspokes = 1,
		  nrings = 4,
		  tmin = 0,
		  tmax = 75,
		  dist = 'uniform'
	       }

   local last_spot
   local ring = 1
   for spot in utils.ringNspoke( pars ) do

      local exp = pars.rmin + ( ring - 1 ) * ( pars.rmax - pars.rmin ) / ( pars.nrings - 1 )
      assert_equal( exp, spot.r, "ring " .. ring )

      ring = ring + 1
      last_spot = spot
   end

   assert_equal( pars.rmax, last_spot.r )

end


